;; Generated package description from landmark.el  -*- no-byte-compile: t -*-
(define-package "landmark" "1.0.0.20201201.211453" "Neural-network robot that learns landmarks" 'nil :url "https://elpa.gnu.org/packages/landmark.html" :keywords '("games" "neural network" "adaptive search" "chemotaxis") :authors '(("Terrence Brannon" . "metaperl@gmail.com")) :maintainer '(nil . "emacs-devel@gnu.org"))
